@@ -45,6 +45,7 @@
 
 #include "mountP.h"
 #include "strutils.h"
+#include "loopdev.h"
 
 /**
  * mnt_new_table:
@@ -210,7 +211,18 @@ int mnt_table_remove_fs(struct libmnt_table *tb, struct libmnt_fs *fs)
  * @tb: mountinfo file (/proc/self/mountinfo)
  * @root: returns pointer to the root filesystem (/)
  *
- * Returns: 0 on success or -1 case of error.
+ * The function uses parent ID from mountinfo file to determine root filesystem
+ * (the filesystem with the smallest ID). The function is designed mostly for
+ * applications where is necessary to sort mountpoints by IDs to get the tree
+ * of the mountpoints (e.g. findmnt default output).
+ *
+ * If you're not sure than use
+ *
+ *	mnt_table_find_target(tb, "/", MNT_ITER_BACKWARD);
+ *
+ * this is more robust and usable for arbitrary tab file (including fstab).
+ *
+ * Returns: 0 on success or less then zero case of error.
  */
 int mnt_table_get_root_fs(struct libmnt_table *tb, struct libmnt_fs **root)
 {
@@ -413,6 +425,53 @@ int mnt_table_set_iter(struct libmnt_table *tb, struct libmnt_iter *itr, struct
 	return 0;
 }
 
+/**
+ * mnt_table_find_mountpoint:
+ * @tb: tab pointer
+ * @path: directory
+ * @direction: MNT_ITER_{FORWARD,BACKWARD}
+ *
+ * Same like mnt_get_mountpoint(), but this function does not rely on
+ * st_dev numbers.
+ *
+ * Returns: a tab entry or NULL.
+ */
+struct libmnt_fs *mnt_table_find_mountpoint(struct libmnt_table *tb,
+					    const char *path,
+					    int direction)
+{
+	char *mnt;
+
+	if (!tb || !path)
+		return NULL;
+	if (direction != MNT_ITER_FORWARD && direction != MNT_ITER_BACKWARD)
+		return NULL;
+
+	DBG(TAB, mnt_debug_h(tb, "lookup MOUNTPOINT: %s", path));
+
+	mnt = strdup(path);
+	if (!mnt)
+		return NULL;
+
+	do {
+		char *p;
+		struct libmnt_fs *fs;
+
+		fs = mnt_table_find_target(tb, mnt, direction);
+		if (fs) {
+			free(mnt);
+			return fs;
+		}
+
+		p = stripoff_last_component(mnt);
+		if (!p || !*p)
+			break;
+	} while (mnt && *(mnt + 1) != '\0');
+
+	free(mnt);
+	return mnt_table_find_target(tb, "/", direction);
+}
+
 /**
  * mnt_table_find_target:
  * @tb: tab pointer
@@ -437,6 +496,8 @@ struct libmnt_fs *mnt_table_find_target(struct libmnt_table *tb, const char *pat
 
 	if (!tb || !path)
 		return NULL;
+	if (direction != MNT_ITER_FORWARD && direction != MNT_ITER_BACKWARD)
+		return NULL;
 
 	DBG(TAB, mnt_debug_h(tb, "lookup TARGET: %s", path));
 
@@ -505,6 +566,10 @@ struct libmnt_fs *mnt_table_find_srcpath(struct libmnt_table *tb, const char *pa
 	const char *p;
 
 	assert(tb);
+	if (!tb || !path)
+		return NULL;
+	if (direction != MNT_ITER_FORWARD && direction != MNT_ITER_BACKWARD)
+		return NULL;
 
 	DBG(TAB, mnt_debug_h(tb, "lookup srcpath: %s", path));
 
@@ -609,6 +674,8 @@ struct libmnt_fs *mnt_table_find_tag(struct libmnt_table *tb, const char *tag,
 
 	if (!tb || !tag || !val)
 		return NULL;
+	if (direction != MNT_ITER_FORWARD && direction != MNT_ITER_BACKWARD)
+		return NULL;
 
 	DBG(TAB, mnt_debug_h(tb, "lookup by TAG: %s %s", tag, val));
 
@@ -651,6 +718,8 @@ struct libmnt_fs *mnt_table_find_source(struct libmnt_table *tb,
 
 	if (!tb)
 		return NULL;
+	if (direction != MNT_ITER_FORWARD && direction != MNT_ITER_BACKWARD)
+		return NULL;
 
 	DBG(TAB, mnt_debug_h(tb, "lookup SOURCE: %s", source));
 
@@ -694,6 +763,8 @@ struct libmnt_fs *mnt_table_find_pair(struct libmnt_table *tb, const char *sourc
 
 	if (!tb || !target)
 		return NULL;
+	if (direction != MNT_ITER_FORWARD && direction != MNT_ITER_BACKWARD)
+		return NULL;
 
 	DBG(TAB, mnt_debug_h(tb, "lookup SOURCE: %s TARGET: %s", source, target));
 
@@ -729,6 +800,8 @@ struct libmnt_fs *mnt_table_find_devno(struct libmnt_table *tb,
 
 	if (!tb)
 		return NULL;
+	if (direction != MNT_ITER_FORWARD && direction != MNT_ITER_BACKWARD)
+		return NULL;
 
 	DBG(TAB, mnt_debug_h(tb, "lookup DEVNO: %d", (int) devno));
 
@@ -889,12 +962,13 @@ static int is_mountinfo(struct libmnt_table *tb)
  *
  * This function is designed mostly for "mount -a".
  *
- * TODO: check for loopdev (see mount/mount.c is_fstab_entry_mounted().
- *
  * Returns: 0 or 1
  */
 int mnt_table_is_fs_mounted(struct libmnt_table *tb, struct libmnt_fs *fstab_fs)
 {
+	struct libmnt_iter itr;
+	struct libmnt_fs *fs;
+
 	char *root = NULL;
 	const char *src = NULL, *tgt = NULL;
 	char *xtgt = NULL;
@@ -903,20 +977,25 @@ int mnt_table_is_fs_mounted(struct libmnt_table *tb, struct libmnt_fs *fstab_fs)
 	assert(tb);
 	assert(fstab_fs);
 
-	if (mnt_fs_is_swaparea(fstab_fs) || mnt_table_get_nents(tb) == 0)
+	DBG(FS, mnt_debug_h(fstab_fs, "is FS mounted? [target=%s]",
+				mnt_fs_get_target(fstab_fs)));
+
+	if (mnt_fs_is_swaparea(fstab_fs) || mnt_table_get_nents(tb) == 0) {
+		DBG(FS, mnt_debug_h(fstab_fs, "- ignore (swap or no data)"));
 		return 0;
+	}
 
 	if (is_mountinfo(tb)) {
 		/* @tb is mountinfo, so we can try to use fs-roots */
-		struct libmnt_fs *fs;
+		struct libmnt_fs *rootfs;
 		int flags = 0;
 
 		if (mnt_fs_get_option(fstab_fs, "bind", NULL, NULL) == 0)
 			flags = MS_BIND;
 
-		fs = mnt_table_get_fs_root(tb, fstab_fs, flags, &root);
-		if (fs)
-			src = mnt_fs_get_srcpath(fs);
+		rootfs = mnt_table_get_fs_root(tb, fstab_fs, flags, &root);
+		if (rootfs)
+			src = mnt_fs_get_srcpath(rootfs);
 	}
 
 	if (!src)
@@ -927,47 +1006,71 @@ int mnt_table_is_fs_mounted(struct libmnt_table *tb, struct libmnt_fs *fstab_fs)
 
 	tgt = mnt_fs_get_target(fstab_fs);
 
-	if (tgt && src) {
-		struct libmnt_iter itr;
-		struct libmnt_fs *fs;
+	if (!tgt || !src) {
+		DBG(FS, mnt_debug_h(fstab_fs, "- ignore (no source/target)"));
+		goto done;
+	}
+	mnt_reset_iter(&itr, MNT_ITER_FORWARD);
 
-		mnt_reset_iter(&itr, MNT_ITER_FORWARD);
+	while (mnt_table_next_fs(tb, &itr, &fs) == 0) {
 
-		while (mnt_table_next_fs(tb, &itr, &fs) == 0) {
-
-			if (!mnt_fs_streq_srcpath(fs, src))
-				continue;
-
-			if (root) {
-				const char *r = mnt_fs_get_root(fs);
-				if (!r || strcmp(r, root) != 0)
-					continue;
-			}
-
-			/*
-			 * Compare target, try to minimize number of situations
-			 * when we need to canonicalize the path to avoid
-			 * readlink() on mountpoints.
+		if (!mnt_fs_streq_srcpath(fs, src)) {
+			/* The source does not match. Maybe the source is a loop
+			 * device backing file.
 			 */
-			if (!xtgt) {
-				if (mnt_fs_streq_target(fs, tgt))
-					break;
-				if (tb->cache)
-					xtgt = mnt_resolve_path(tgt, tb->cache);
-			}
-			if (xtgt && mnt_fs_streq_target(fs, xtgt))
-				break;
+			uint64_t offset = 0;
+			char *val;
+			size_t len;
+			int flags;
 
+			if (!mnt_fs_is_kernel(fs) ||
+			    !mnt_fs_get_srcpath(fs) ||
+			    !startswith(mnt_fs_get_srcpath(fs), "/dev/loop"))
+				continue;	/* does not look like loopdev */
+
+			if (mnt_fs_get_option(fstab_fs, "offset", &val, &len) == 0 &&
+			    mnt_parse_offset(val, len, &offset)) {
+				DBG(FS, mnt_debug_h(fstab_fs, "failed to parse offset="));
+				continue;
+			} else
+				flags = LOOPDEV_FL_OFFSET;
+
+			if (loopdev_is_used(mnt_fs_get_srcpath(fs), src, offset, flags))
+				break;
 		}
-		if (fs)
-			rc = 1;		/* success */
+
+		if (root) {
+			const char *r = mnt_fs_get_root(fs);
+			if (!r || strcmp(r, root) != 0)
+				continue;
+		}
+
+		/*
+		 * Compare target, try to minimize number of situations when we
+		 * need to canonicalize the path to avoid readlink() on
+		 * mountpoints.
+		 */
+		if (!xtgt) {
+			if (mnt_fs_streq_target(fs, tgt))
+				break;
+			if (tb->cache)
+				xtgt = mnt_resolve_path(tgt, tb->cache);
+		}
+		if (xtgt && mnt_fs_streq_target(fs, xtgt))
+			break;
 	}
 
+	if (fs)
+		rc = 1;		/* success */
+done:
 	free(root);
+
+	DBG(TAB, mnt_debug_h(tb, "mnt_table_is_fs_mounted: %s [rc=%d]", src, rc));
 	return rc;
 }
 
 #ifdef TEST_PROGRAM
+#include "pathnames.h"
 
 static int parser_errcb(struct libmnt_table *tb, const char *filename, int line)
 {
@@ -1130,6 +1233,33 @@ done:
 	return rc;
 }
 
+int test_find_mountpoint(struct libmnt_test *ts, int argc, char *argv[])
+{
+	struct libmnt_table *tb;
+	struct libmnt_fs *fs;
+	struct libmnt_cache *mpc = NULL;
+	int rc = -1;
+
+	tb = mnt_new_table_from_file(_PATH_PROC_MOUNTINFO);
+	if (!tb)
+		return -1;
+	mpc = mnt_new_cache();
+	if (!mpc)
+		goto done;
+	mnt_table_set_cache(tb, mpc);
+
+	fs = mnt_table_find_mountpoint(tb, argv[1], MNT_ITER_BACKWARD);
+	if (!fs)
+		goto done;
+
+	mnt_fs_print_debug(fs, stdout);
+	rc = 0;
+done:
+	mnt_free_table(tb);
+	mnt_free_cache(mpc);
+	return rc;
+}
+
 static int test_is_mounted(struct libmnt_test *ts, int argc, char *argv[])
 {
 	struct libmnt_table *tb = NULL, *fstab = NULL;
@@ -1184,6 +1314,7 @@ int main(int argc, char *argv[])
 	{ "--find-forward",  test_find_fw, "<file> <source|target> <string>" },
 	{ "--find-backward", test_find_bw, "<file> <source|target> <string>" },
 	{ "--find-pair",     test_find_pair, "<file> <source> <target>" },
+	{ "--find-mountpoint", test_find_mountpoint, "<path>" },
 	{ "--copy-fs",       test_copy_fs, "<file>  copy root FS from the file" },
 	{ "--is-mounted",    test_is_mounted, "<fstab> check what from <file> are already mounted" },
 	{ NULL }
