@@ -23,6 +23,7 @@
 #include <fcntl.h>
 
 #include "c.h"
+#include "colors.h"
 #include "nls.h"
 #include "strutils.h"
 #include "xalloc.h"
@@ -32,6 +33,7 @@
 #include "closestream.h"
 #include "optutils.h"
 #include "mangle.h"
+#include "pager.h"
 
 /* Close the log.  Currently a NOP. */
 #define SYSLOG_ACTION_CLOSE          0
@@ -56,6 +58,18 @@
 /* Return size of the log buffer */
 #define SYSLOG_ACTION_SIZE_BUFFER   10
 
+/*
+ * Colors
+ */
+#define DMESG_COLOR_SUBSYS	UL_COLOR_BROWN
+#define DMESG_COLOR_TIME	UL_COLOR_GREEN
+#define DMESG_COLOR_RELTIME	UL_COLOR_BOLD_GREEN
+#define DMESG_COLOR_ALERT	UL_COLOR_REVERSE UL_COLOR_RED
+#define DMESG_COLOR_CRIT	UL_COLOR_BOLD_RED
+#define DMESG_COLOR_ERR		UL_COLOR_RED
+#define DMESG_COLOR_WARN	UL_COLOR_BOLD
+#define DMESG_COLOR_SEGFAULT	UL_COLOR_HALFBRIGHT UL_COLOR_RED
+
 /*
  * Priority and facility names
  */
@@ -147,7 +161,9 @@ struct dmesg_control {
 			notime:1,	/* don't print timestamp */
 			delta:1,	/* show time deltas */
 			reltime:1,	/* show human readable relative times */
-			ctime:1;	/* show human readable time */
+			ctime:1,	/* show human readable time */
+			pager:1,	/* pipe output into a pager */
+			color:1;	/* colorize messages */
 };
 
 struct dmesg_record {
@@ -173,55 +189,82 @@ struct dmesg_record {
 
 static int read_kmsg(struct dmesg_control *ctl);
 
+static int set_level_color(int log_level, const char *mesg, size_t mesgsz)
+{
+	switch (log_level) {
+	case LOG_ALERT:
+		color_enable(DMESG_COLOR_ALERT);
+		return 0;
+	case LOG_CRIT:
+		color_enable(DMESG_COLOR_CRIT);
+		return 0;
+	case LOG_ERR:
+		color_enable(DMESG_COLOR_ERR);
+		return 0;
+	case LOG_WARNING:
+		color_enable(DMESG_COLOR_WARN);
+		return 0;
+	default:
+		break;
+	}
+
+	/* well, sometimes the messges contains important keywords, but in
+	 * non-warning/error messages
+	 */
+	if (memmem(mesg, mesgsz, "segfault at", 11)) {
+		color_enable(DMESG_COLOR_SEGFAULT);
+		return 0;
+	}
+
+	return 1;
+}
 
 static void __attribute__((__noreturn__)) usage(FILE *out)
 {
 	size_t i;
 
-	fputs(_("\nUsage:\n"), out);
-	fprintf(out,
-	      _(" %s [options]\n"), program_invocation_short_name);
-
-	fputs(_("\nOptions:\n"), out);
-	fputs(_(" -C, --clear                 clear the kernel ring buffer\n"
-		" -c, --read-clear            read and clear all messages\n"
-		" -D, --console-off           disable printing messages to console\n"
-		" -d, --show-delta            show time delta between printed messages\n"
-		" -e, --reltime               show local time and time delta in readable format\n"
-		" -E, --console-on            enable printing messages to console\n"
-		" -F, --file <file>           use the file instead of the kernel log buffer\n"
-		" -f, --facility <list>       restrict output to defined facilities\n"
-		" -h, --help                  display this help and exit\n"
-		" -k, --kernel                display kernel messages\n"
-		" -l, --level <list>          restrict output to defined levels\n"
-		" -n, --console-level <level> set level of messages printed to console\n"
-		" -r, --raw                   print the raw message buffer\n"
-		" -S, --syslog                force to use syslog(2) rather than /dev/kmsg\n"
-		" -s, --buffer-size <size>    buffer size to query the kernel ring buffer\n"
-		" -T, --ctime                 show human readable timestamp (could be \n"
-		"                             inaccurate if you have used SUSPEND/RESUME)\n"
-		" -t, --notime                don't print messages timestamp\n"
-		" -u, --userspace             display userspace messages\n"
-		" -V, --version               output version information and exit\n"
-		" -w, --follow                wait for new messages\n"
-		" -x, --decode                decode facility and level to readable string\n"), out);
-
+	fputs(USAGE_HEADER, out);
+	fprintf(out, _(" %s [options]\n"), program_invocation_short_name);
+	fputs(USAGE_OPTIONS, out);
+	fputs(_(" -C, --clear                 clear the kernel ring buffer\n"), out);
+	fputs(_(" -c, --read-clear            read and clear all messages\n"), out);
+	fputs(_(" -D, --console-off           disable printing messages to console\n"), out);
+	fputs(_(" -d, --show-delta            show time delta between printed messages\n"), out);
+	fputs(_(" -e, --reltime               show local time and time delta in readable format\n"), out);
+	fputs(_(" -E, --console-on            enable printing messages to console\n"), out);
+	fputs(_(" -F, --file <file>           use the file instead of the kernel log buffer\n"), out);
+	fputs(_(" -f, --facility <list>       restrict output to defined facilities\n"), out);
+	fputs(_(" -H, --human                 human readable output\n"), out);
+	fputs(_(" -k, --kernel                display kernel messages\n"), out);
+	fputs(_(" -L, --color                 colorize messages\n"), out);
+	fputs(_(" -l, --level <list>          restrict output to defined levels\n"), out);
+	fputs(_(" -n, --console-level <level> set level of messages printed to console\n"), out);
+	fputs(_(" -P, --nopager               do not pipe output into a pager\n"), out);
+	fputs(_(" -r, --raw                   print the raw message buffer\n"), out);
+	fputs(_(" -S, --syslog                force to use syslog(2) rather than /dev/kmsg\n"), out);
+	fputs(_(" -s, --buffer-size <size>    buffer size to query the kernel ring buffer\n"), out);
+	fputs(_(" -T, --ctime                 show human readable timestamp (could be \n"
+		"                               inaccurate if you have used SUSPEND/RESUME)\n"), out);
+	fputs(_(" -t, --notime                don't print messages timestamp\n"), out);
+	fputs(_(" -u, --userspace             display userspace messages\n"), out);
+	fputs(_(" -w, --follow                wait for new messages\n"), out);
+	fputs(_(" -x, --decode                decode facility and level to readable string\n"), out);
+	fputs(USAGE_SEPARATOR, out);
+	fputs(USAGE_HELP, out);
+	fputs(USAGE_VERSION, out);
 	fputs(_("\nSupported log facilities:\n"), out);
-	for (i = 0; i < ARRAY_SIZE(level_names); i++) {
-		fprintf(stderr, " %7s - %s\n",
-				facility_names[i].name,
-				_(facility_names[i].help));
-	}
+	for (i = 0; i < ARRAY_SIZE(level_names); i++)
+		fprintf(out, " %7s - %s\n",
+			facility_names[i].name,
+			_(facility_names[i].help));
 
 	fputs(_("\nSupported log levels (priorities):\n"), out);
-	for (i = 0; i < ARRAY_SIZE(level_names); i++) {
-		fprintf(stderr, " %7s - %s\n",
-				level_names[i].name,
-				_(level_names[i].help));
-	}
-
-	fputc('\n', out);
-
+	for (i = 0; i < ARRAY_SIZE(level_names); i++)
+		fprintf(out, " %7s - %s\n",
+			level_names[i].name,
+			_(level_names[i].help));
+	fputs(USAGE_SEPARATOR, out);
+	fprintf(out, USAGE_MAN_TAIL("dmesg(q)"));
 	exit(out == stderr ? EXIT_FAILURE : EXIT_SUCCESS);
 }
 
@@ -637,7 +680,7 @@ static int get_next_syslog_record(struct dmesg_control *ctl,
 			continue;	/* error or empty line? */
 
 		if (*begin == '<') {
-			if (ctl->fltr_lev || ctl->fltr_fac || ctl->decode)
+			if (ctl->fltr_lev || ctl->fltr_fac || ctl->decode || ctl->color)
 				begin = parse_faclev(begin + 1, &rec->facility,
 						     &rec->level);
 			else
@@ -752,10 +795,32 @@ static double record_count_delta(struct dmesg_control *ctl,
 	return delta;
 }
 
+static const char *get_subsys_delimiter(const char *mesg, size_t mesg_size)
+{
+	const char *p = mesg;
+	size_t sz = mesg_size;
+
+	while (sz > 0) {
+		const char *d = strnchr(p, sz, ':');
+		if (!d)
+			return NULL;
+		sz -= d - p;
+		if (sz) {
+			if (isblank(*(d + 1)))
+				return d;
+			p = d + 1;
+		}
+	}
+	return NULL;
+}
+
 static void print_record(struct dmesg_control *ctl,
 			 struct dmesg_record *rec)
 {
 	char buf[256];
+	int has_color = 0;
+	const char *mesg;
+	size_t mesg_size;
 
 	if (!accept_record(ctl, rec))
 		return;
@@ -781,7 +846,9 @@ static void print_record(struct dmesg_control *ctl,
 	/*
 	 * facility : priority :
 	 */
-	if (ctl->decode && rec->level >= 0 && rec->facility >= 0)
+	if (ctl->decode &&
+	    -1 < rec->level    && rec->level     < (int) ARRAY_SIZE(level_names) &&
+	    -1 < rec->facility && rec->facility  < (int) ARRAY_SIZE(facility_names))
 		printf("%-6s:%-6s: ", facility_names[rec->facility].name,
 				      level_names[rec->level].name);
 
@@ -789,6 +856,8 @@ static void print_record(struct dmesg_control *ctl,
 	 * [sec.usec <delta>] or [ctime <delta>]
 	 */
 	if (ctl->delta) {
+		if (ctl->color)
+			color_enable(DMESG_COLOR_TIME);
 		if (ctl->ctime)
 			printf("[%s ", record_ctime(ctl, rec, buf, sizeof(buf)));
 		else if (ctl->notime)
@@ -797,12 +866,19 @@ static void print_record(struct dmesg_control *ctl,
 			printf("[%5d.%06d ", (int) rec->tv.tv_sec,
 					     (int) rec->tv.tv_usec);
 		printf("<%12.06f>] ", record_count_delta(ctl, rec));
+		if (ctl->color)
+			color_disable();
 
 	/*
 	 * [ctime]
 	 */
-	} else if (ctl->ctime)
+	} else if (ctl->ctime) {
+		if (ctl->color)
+			color_enable(DMESG_COLOR_TIME);
 		printf("[%s] ", record_ctime(ctl, rec, buf, sizeof(buf)));
+		if (ctl->color)
+			color_disable();
+	}
 
 	/*
 	 * [reltime]
@@ -816,13 +892,20 @@ static void print_record(struct dmesg_control *ctl,
 
 		if (cur.tm_min  != ctl->lasttm.tm_min ||
 		    cur.tm_hour != ctl->lasttm.tm_hour ||
-		    cur.tm_yday != ctl->lasttm.tm_yday)
+		    cur.tm_yday != ctl->lasttm.tm_yday) {
+			if (ctl->color)
+				color_enable(DMESG_COLOR_RELTIME);
 			printf("[%s] ", short_ctime(&cur, buf, sizeof(buf)));
-		else if (delta < 10)
-			printf("[  %+8.06f] ", delta);
-		else
-			printf("[ %+9.06f] ", delta);
-
+		} else {
+			if (ctl->color)
+				color_enable(DMESG_COLOR_TIME);
+			if (delta < 10)
+				printf("[  %+8.06f] ", delta);
+			else
+				printf("[ %+9.06f] ", delta);
+		}
+		if (ctl->color)
+			color_disable();
 		ctl->lasttm = cur;
 	}
 
@@ -835,13 +918,39 @@ static void print_record(struct dmesg_control *ctl,
 	 * the [sec.usec] string.
 	 */
 	if (ctl->method == DMESG_METHOD_KMSG &&
-	    !ctl->notime && !ctl->delta && !ctl->ctime && !ctl->reltime)
+	    !ctl->notime && !ctl->delta && !ctl->ctime && !ctl->reltime) {
+		if (ctl->color)
+			color_enable(DMESG_COLOR_TIME);
 		printf("[%5d.%06d] ", (int) rec->tv.tv_sec, (int) rec->tv.tv_usec);
+		if (ctl->color)
+			color_disable();
+	}
 
 mesg:
-	safe_fwrite(rec->mesg, rec->mesg_size, stdout);
+	mesg = rec->mesg;
+	mesg_size = rec->mesg_size;
 
-	if (*(rec->mesg + rec->mesg_size - 1) != '\n')
+	/* Colorize output */
+	if (ctl->color) {
+		/* subsystem prefix */
+		const char *subsys = get_subsys_delimiter(mesg, mesg_size);
+		if (subsys) {
+			color_enable(DMESG_COLOR_SUBSYS);
+			safe_fwrite(mesg, subsys - mesg, stdout);
+			color_disable();
+
+			mesg_size -= subsys - mesg;
+			mesg = subsys;
+		}
+		/* error, alert .. etc. colors */
+		has_color = set_level_color(rec->level, mesg, mesg_size) == 0;
+		safe_fwrite(mesg, mesg_size, stdout);
+		if (has_color)
+			color_disable();
+	} else
+		safe_fwrite(mesg, mesg_size, stdout);
+
+	if (*(mesg + mesg_size - 1) != '\n')
 		putchar('\n');
 }
 
@@ -942,7 +1051,8 @@ static int parse_kmsg_record(struct dmesg_control *ctl,
 		p++;
 
 	/* A) priority and facility */
-	if (ctl->fltr_lev || ctl->fltr_fac || ctl->decode || ctl->raw)
+	if (ctl->fltr_lev || ctl->fltr_fac || ctl->decode ||
+	    ctl->raw || ctl->color)
 		p = parse_faclev(p, &rec->facility, &rec->level);
 	else
 		p = skip_item(p, end, ",");
@@ -1026,13 +1136,10 @@ static int read_kmsg(struct dmesg_control *ctl)
 	return 0;
 }
 
-#define EXCL_ACT_ERR "--{clear,read-clear,console-level,console-on,console-off}"
-#define EXCL_SYS_ERR "--{syslog,follow}"
-
 int main(int argc, char *argv[])
 {
 	char *buf = NULL;
-	int  c;
+	int  c, nopager = 0;
 	int  console_level = 0;
 	int  klog_rc = 0;
 	ssize_t n;
@@ -1046,6 +1153,7 @@ int main(int argc, char *argv[])
 	static const struct option longopts[] = {
 		{ "buffer-size",   required_argument, NULL, 's' },
 		{ "clear",         no_argument,	      NULL, 'C' },
+		{ "color",         no_argument,	      NULL, 'L' },
 		{ "console-level", required_argument, NULL, 'n' },
 		{ "console-off",   no_argument,       NULL, 'D' },
 		{ "console-on",    no_argument,       NULL, 'E' },
@@ -1053,6 +1161,7 @@ int main(int argc, char *argv[])
 		{ "file",          required_argument, NULL, 'F' },
 		{ "facility",      required_argument, NULL, 'f' },
 		{ "follow",        no_argument,       NULL, 'w' },
+		{ "human",         no_argument,       NULL, 'H' },
 		{ "help",          no_argument,	      NULL, 'h' },
 		{ "kernel",        no_argument,       NULL, 'k' },
 		{ "level",         required_argument, NULL, 'l' },
@@ -1063,6 +1172,7 @@ int main(int argc, char *argv[])
 		{ "show-delta",    no_argument,	      NULL, 'd' },
 		{ "ctime",         no_argument,       NULL, 'T' },
 		{ "notime",        no_argument,       NULL, 't' },
+		{ "nopager",       no_argument,       NULL, 'P' },
 		{ "userspace",     no_argument,       NULL, 'u' },
 		{ "version",       no_argument,	      NULL, 'V' },
 		{ NULL,	           0, NULL, 0 }
@@ -1070,6 +1180,8 @@ int main(int argc, char *argv[])
 
 	static const ul_excl_t excl[] = {	/* rows and cols in in ASCII order */
 		{ 'C','D','E','c','n' },	/* clear,off,on,read-clear,level*/
+		{ 'H','r' },			/* human, raw */
+		{ 'L','r' },			/* color, raw */
 		{ 'S','w' },			/* syslog,follow */
 		{ 0 }
 	};
@@ -1080,7 +1192,7 @@ int main(int argc, char *argv[])
 	textdomain(PACKAGE);
 	atexit(close_stdout);
 
-	while ((c = getopt_long(argc, argv, "CcDdEeF:f:hkl:n:rSs:TtuVwx",
+	while ((c = getopt_long(argc, argv, "CcDdEeF:f:HhkLl:n:iPrSs:TtuVwx",
 				longopts, NULL)) != -1) {
 
 		err_exclusive_options(c, longopts, excl, excl_st);
@@ -1102,9 +1214,7 @@ int main(int argc, char *argv[])
 			ctl.action = SYSLOG_ACTION_CONSOLE_ON;
 			break;
 		case 'e':
-			ctl.boot_time = get_boot_time();
-			if (ctl.boot_time)
-				ctl.reltime = 1;
+			ctl.reltime = 1;
 			break;
 		case 'F':
 			ctl.filename = optarg;
@@ -1116,6 +1226,11 @@ int main(int argc, char *argv[])
 					     ctl.facilities, parse_facility) < 0)
 				return EXIT_FAILURE;
 			break;
+		case 'H':
+			ctl.reltime = 1;
+			ctl.color = 1;
+			ctl.pager = 1;
+			break;
 		case 'h':
 			usage(stdout);
 			break;
@@ -1123,6 +1238,9 @@ int main(int argc, char *argv[])
 			ctl.fltr_fac = 1;
 			setbit(ctl.facilities, FAC_BASE(LOG_KERN));
 			break;
+		case 'L':
+			ctl.color = 1;
+			break;
 		case 'l':
 			ctl.fltr_lev= 1;
 			if (string_to_bitarray(optarg,
@@ -1133,6 +1251,9 @@ int main(int argc, char *argv[])
 			ctl.action = SYSLOG_ACTION_CONSOLE_LEVEL;
 			console_level = parse_level(optarg, 0);
 			break;
+		case 'P':
+			nopager = 1;
+			break;
 		case 'r':
 			ctl.raw = 1;
 			break;
@@ -1189,12 +1310,25 @@ int main(int argc, char *argv[])
 	if (ctl.reltime && ctl.ctime)
 		errx(EXIT_FAILURE, _("--reltime can't be used together with --ctime "));
 
+	if (ctl.reltime) {
+		ctl.boot_time = get_boot_time();
+		if (!ctl.boot_time)
+			ctl.reltime = 0;
+	}
+	if (ctl.color)
+		ctl.color = colors_init() ? 1 : 0;
+
+	ctl.pager = nopager ? 0 : ctl.pager;
+	if (ctl.pager)
+		setup_pager();
+
 	switch (ctl.action) {
 	case SYSLOG_ACTION_READ_ALL:
 	case SYSLOG_ACTION_READ_CLEAR:
 		if (ctl.method == DMESG_METHOD_KMSG && init_kmsg(&ctl) != 0)
 			ctl.method = DMESG_METHOD_SYSLOG;
-
+		if (ctl.pager)
+			setup_pager();
 		n = read_buffer(&ctl, &buf);
 		if (n > 0)
 			print_buffer(&ctl, buf, n);
